/*
 * EKG2_v2.c
 *
 * Created: 26.07.2014 14:34:08
 *  Author: landreas
 */ 

/* DEFINES, MACROS AND CONSTANTS*/

//Frequency defined for the delay 
#define F_CPU 1000000UL

//A couple of macros for the DMA. 
#define DMA_off() DMA.CTRL &= ~DMA_ENABLE_bm

//Buffer size of the DMA channels. 1024 bytes = 512 samples / second, since a sample is 2 bytes. 
#define BUFFER_SIZE 1024

//Number of times the user can save a timestamp in a session
#define BUTTONCOUNT_LIMIT 50	


/* INCLUDES */
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "board.h"
#include "ff.h"



/*GLOBAL VARIABLES*/

//Variables for the FatFs file system
FATFS FatFs; 
FIL Fil;			

//Function pointer for the states of the state machine.
uint8_t (*function_pointer)(void);

//Buffer in the SRAM to temporarily store the data for the DMA channels
char buffer_a[BUFFER_SIZE], buffer_b[BUFFER_SIZE];

//Variables for the timestamp system.
volatile uint32_t timestamp = 0;
uint32_t buttonstamps[BUTTONCOUNT_LIMIT];
uint8_t buttoncounter = 0;


/* PROTOTYPES */
uint8_t sdcard_write(char*, uint16_t);
uint8_t sdcard_init(void);
void sdcard_close(void);
void log_ecgdata(void);
void sleep(void);
void die(void);
void wake(void);
void rtc_init(void);
void tosc_init(void);
void dma_init(void);
void pmic_init(void);
void sleepmgr_sleep(void);



int main(void)
{
	//initialize board
	board_init();
	//enable interrupts
	sei();
	//starting in the sleep state
	function_pointer = &sleep;
    while(1)
    {
		function_pointer();
        //TODO:: Please write your application code 
    }
}  

void die(void){
	
	//SHUTDOWN
	GLED_off();
	sdcard_close();
	DMA_off();
	
	//Blink red LED until the error is fixed.
	while(function_pointer == die){
		_delay_ms(1000);
		RLED_toggle();
	}
}

//Initialize the Programmable Multilevel Interrupt Controller
void pmic_init(void){
	//Enable all interrupts
	PMIC.CTRL = PMIC_MEDLVLEN_bm | PMIC_HILVLEN_bm | PMIC_LOLVLEN_bm;
}





void adc_init(void){
	
	//Turn on power to the ADC in case it has been disabled
	PR.PRPA &= ~PR_ADC_bm;
	
	//Select the voltage on AREFA as reference voltage. The measured voltage must be below this voltage for correct measurement.
	//The reference voltage cannot be AVCC for XMega devices. AREFA must be 0.6V below VCC for correct operation. 
	ADCA.REFCTRL = ADC_REFSEL_AREFA_gc;
	//Turn on prescaler. clk_adc should be between 100kHz and 2MHz. F_CPU is 0.5MHz so prescale_4 = 125kHz.
	ADCA.PRESCALER = ADC_PRESCALER_DIV4_gc;
	//Clear intflags
	ADCA.INTFLAGS = 0x0F;
	//Self explanatory
	ADCA.CH0.CTRL = ADC_CH_INPUTMODE_SINGLEENDED_gc | ADC_CH_GAIN_1X_gc;
	//PA5 as input for channel 0
	ADCA.CH0.MUXCTRL = SIGNAL_MUX_PIN;
	//Wait for ADC to initialize. 100ms is just an arbitrary number. 
	_delay_ms(100);
	//Event channel 0 triggers ADC channel 0
	ADCA.EVCTRL = ADC_SWEEP_0_gc | ADC_EVSEL_0123_gc | ADC_EVACT_CH0_gc; 
	//Lastly, turn on the ADC
	ADCA.CTRLA |= ADC_ENABLE_bm;
	
}

//Initialize all onboard functions
void board_init(void){
	
	sysclk_init();
	adc_init();
	dma_init();
	button_init();
	charger_init();
	rtc_init();
	tosc_init();
	pmic_init(); 
	
	GLED_init();
	RLED_init();
	ENABLE_init();
	SIGNAL_init();
}

void dma_init(void){
	
	//Turn on power to the Event System and DMAC
	PR.PRGEN &= ~PR_EVSYS_bm;
	PR.PRGEN &= ~PR_DMA_bm;
		
	EVSYS.CH0MUX = EVSYS_CHMUX_RTC_OVF_gc; // trigger on timer overflow

	// reset DMA controller
	DMA.CTRL = 0;
	DMA.CTRL = DMA_RESET_bm;
	while ((DMA.CTRL & DMA_RESET_bm) != 0)
	;
	// configure DMA controller
	DMA.CTRL = DMA_CH_ENABLE_bm | DMA_DBUFMODE_CH01_gc; // double buffered with channels 0 and 1
		
	// channel 0
	// **** TODO: reset dma channels
	DMA.CH0.REPCNT = 0;
	DMA.CH0.CTRLA = DMA_CH_BURSTLEN_2BYTE_gc | DMA_CH_SINGLE_bm | DMA_CH_REPEAT_bm; // ADC result is 2 byte 12 bit word
	DMA.CH0.ADDRCTRL = DMA_CH_SRCRELOAD_BURST_gc | DMA_CH_SRCDIR_INC_gc | // reload source after every burst
	DMA_CH_DESTRELOAD_TRANSACTION_gc | DMA_CH_DESTDIR_INC_gc; // reload dest after every transaction
	DMA.CH0.TRIGSRC = DMA_CH_TRIGSRC_ADCA_CH0_gc;
	DMA.CH0.TRFCNT = BUFFER_SIZE; // always the number of bytes, even if burst length > 1
	DMA.CH0.DESTADDR0 = (( (uint16_t) buffer_a) >> 0) & 0xFF;
	DMA.CH0.DESTADDR1 = (( (uint16_t) buffer_a) >> 8) & 0xFF;
	DMA.CH0.DESTADDR2 = 0;
	DMA.CH0.SRCADDR0 = (( (uint16_t) &ADCA.CH0.RES) >> 0) & 0xFF;
	DMA.CH0.SRCADDR1 = (( (uint16_t) &ADCA.CH0.RES) >> 8) & 0xFF;
	DMA.CH0.SRCADDR2 = 0;
	DMA.CH0.CTRLB |= DMA_CH_TRNINTLVL_HI_gc;

	// channel 1
	DMA.CH1.REPCNT = 0;
	DMA.CH1.CTRLA = DMA_CH_BURSTLEN_2BYTE_gc | DMA_CH_SINGLE_bm | DMA_CH_REPEAT_bm; // ADC result is 2 byte 12 bit word
	DMA.CH1.ADDRCTRL = DMA_CH_SRCRELOAD_BURST_gc | DMA_CH_SRCDIR_INC_gc | // reload source after every burst
	DMA_CH_DESTRELOAD_TRANSACTION_gc | DMA_CH_DESTDIR_INC_gc; // reload dest after every transaction
	DMA.CH1.TRIGSRC = DMA_CH_TRIGSRC_ADCA_CH0_gc;
	DMA.CH1.TRFCNT = BUFFER_SIZE;
	DMA.CH1.DESTADDR0 = (( (uint16_t) buffer_b) >> 0) & 0xFF;
	DMA.CH1.DESTADDR1 = (( (uint16_t) buffer_b) >> 8) & 0xFF;
	DMA.CH1.DESTADDR2 = 0;
	DMA.CH1.SRCADDR0 = (( (uint16_t) &ADCA.CH0.RES) >> 0) & 0xFF;
	DMA.CH1.SRCADDR1 = (( (uint16_t) &ADCA.CH0.RES) >> 8) & 0xFF;
	DMA.CH1.SRCADDR2 = 0;
	DMA.CH1.CTRLB |= DMA_CH_TRNINTLVL_HI_gc;

	DMA.CTRL |= DMA_CH_ENABLE_bm; 
	DMA.CH0.CTRLA |= DMA_CH_ENABLE_bm;
}


//Push button interrupt vector
ISR(PORTB_INT0_vect){
	//Clear interrupt flag
	PORTB.INTFLAGS = PORT_INT0IF_bm;
	//Don't respond if connected and charging
	if(!charging()){
		uint8_t i = 0;
		//Soft switch check
		//Button must be held x seconds to turn on / off
		//Otherwise, a timestamp will be saved while logging
		while(!(BUTTON_PORT.IN & BUTTON_PIN) && i<255){
			for(int j = 0; j < 500; j++);
			i++;
		}
		
		//Check duration and state
		if(i < 200 && function_pointer == log_ecgdata){
			//Timestamp will be saved
			if(buttoncounter<BUTTONCOUNT_LIMIT){
				buttonstamps[buttoncounter] = timestamp;
				buttoncounter++;
			}
			}else if(i >= 200){
			//Soft switch, selecting state
			if(function_pointer == sleep)
				function_pointer = &wake;
			else if(function_pointer == log_ecgdata || function_pointer == die)
				function_pointer = &sleep;
			else
				function_pointer = &sleep;
		}
	}
}

ISR(DMA_CH0_vect){
	//Clear interrupt flag
	DMA.INTFLAGS = DMA_CH0TRNIF_bm;
	//Increase timestamp index
	timestamp++;

	GLED_on();
	
	//Write the ADC/DMA buffer to the memory card
	if(!sdcard_write(buffer_a, BUFFER_SIZE))
	function_pointer = &die; 

	GLED_off();
}

ISR(DMA_CH1_vect){
	//Clear the interrupt flag
	DMA.INTFLAGS = DMA_CH1TRNIF_bm;
	timestamp++;
	GLED_on();
	//char test[7] = {51,52,53,54,55,13,10,0};
	if(!sdcard_write(buffer_b, BUFFER_SIZE))
	function_pointer = &die; //1023 is test. replace with adcverdi.
	GLED_off();
}

ISR(PORTD_INT0_vect){
	if(charging()){
		RLED_on();
		function_pointer = &sleep;
		}else{
		function_pointer = &sleep;
	}
}

void tosc_init(void){
	
	//Select the crystal at TOSC1 and TOSC2 as source with no prescaler and enable it.
	//The logger should have been working with the RTC prescaled to 1024Hz, but simply doesn't.
	CLK.RTCCTRL = CLK_RTCSRC_TOSC32_gc | CLK_RTCEN_bm;
	//Wait for the oscillator to stabilize:
	for(int i = 1; i < 16000; i++);	
}

void rtc_init(void)
{
	//Turn on RTC power
	PR.PRGEN &= ~PR_RTC_bm;
	//The counter will reset upon reaching the periodical register (PER) and trigger an overflow.
	//The overflow will be channeled through the event system and trigger the ADC.
	//The RTC will be prescaled down to 2048Hz.
	//A value of 3 in the PER register means 4 counts. 2048Hz / 4 = 512 Hz which is our sampling rate.
	RTC.PER = 3; 
	RTC.CNT = 0;
	//32768Hz / 16 = 2048Hz and start the RTC clock. 
	RTC.CTRL = RTC_PRESCALER_DIV16_gc;
}

void sysclk_init(void){
	
	//Turn off all modules to save power.
	//The modules will individually be re-enabled during init.
	uint8_t *reg = (uint8_t *)&PR.PRGEN;
	for (int i = 0; i <= 6; i++) {
		*(reg++) = 0xff;
	}
	
	//Tell the CPU that we are going to access a write protected IO register
	CCP = CCP_IOREG_gc;
	//Set the prescaler for the default 2MHz system clock. 
	CLK.PSCTRL = SYSCLK_DIV;
	
}

uint8_t sdcard_init(void){
	
	//Enable the analogue module
	ENABLE_on();
	//Mount the memory card and create a file
	f_mount(&FatFs, "", 0);		/* Give a work area to the default drive */
	//Delete existing file
	f_unlink("newfile.txt");
	
	if (f_open(&Fil, "newfile.txt", FA_WRITE | FA_CREATE_ALWAYS) == FR_OK) 	/* Create a file */
	return 1;
	else
	return 0;
	
}

uint8_t sdcard_write(char *charbuffer, uint16_t size){
	
	UINT bw;
	cli();
	//Write the data to the file
	f_write(&Fil, charbuffer, size, &bw);
	sei();
	//bw = bytes written
	if (bw == size)
	return 1;
	else
	return 0;
	
}

void sdcard_close(void){
	
	//Disable the analogue module
	ENABLE_off();
	//Write saved timestamps to the end of the file
	char stopbytes[2] = {0xFF, 0xFF};
	sdcard_write(stopbytes, 2);
	while(buttoncounter){
		//We better store the timestamps as 2 x uint16 little endian so Matlab can easily read the values along with the samples. 
		//Endianess of the total uint32 doesn't matter. 
		char timestamp_buffer[4] = {buttonstamps[buttoncounter-1]>>16&0xFF, buttonstamps[buttoncounter-1]>>24 & 0xFF,buttonstamps[buttoncounter-1]&0xFF,buttonstamps[buttoncounter-1]>>8&0xFF};
		buttoncounter--;
		sdcard_write(timestamp_buffer, 4);
	}
	//Close and sync the file. Must be done before powerdown or removal of sd card.
	f_close(&Fil);
}

void log_ecgdata(void){
	//Sleep in between the sampling, but keep the state
	sleepmgr_sleep();
}

void sleep(void){
	//Turn off the red 
	if(!charging())
		RLED_off();
	GLED_off();
	DMA_off();
	sdcard_close();
	sleepmgr_sleep();
}

void wake(void){
	//Reset the timestamp system
	//the values inside 'buttonstamps' will be overwritten
	timestamp = 0;
	buttoncounter = 0;
	//The red LED should not be on here
	RLED_off();
	//Initialize the SD card
	if(!sdcard_init())
		function_pointer = &die;
	//The DMA address registries must be reset, so we might as well just call the init function
	dma_init();
	//Enter the logging state
	function_pointer = &log_ecgdata;
}

DWORD get_fattime (void)
{
	/* Returns current time packed into a DWORD variable */
	return	  ((DWORD)(2013 - 1980) << 25)	/* Year 2013 */
	| ((DWORD)7 << 21)				/* Month 7 */
	| ((DWORD)28 << 16)				/* Mday 28 */
	| ((DWORD)0 << 11)				/* Hour 0 */
	| ((DWORD)0 << 5)				/* Min 0 */
	| ((DWORD)0 >> 1);				/* Sec 0 */
}

void sleepmgr_sleep(void){
	
	//Set sleep mode
	SLEEP.CTRL = SLEEP_MODE;
	//Enable sleep mode
	SLEEP.CTRL |= SLEEP_SEN_bm;
	
	//Send the sleep instruction
	__asm__ __volatile__ ( "sleep" "\n\t" :: );
	 
	//Disable sleep mode after CPU has woken up
	SLEEP.CTRL &= ~SLEEP_SEN_bm;
	
}